@@ -2,25 +2,36 @@
 ''' Tests for the BaseRequest and BaseResponse objects and their subclasses. '''
 
 import unittest
-import sys, os.path
+import sys
 import bottle
-from bottle import request, response, tob, touni, tonat, json_dumps, _e
+from bottle import request, tob, touni, tonat, json_dumps, _e, HTTPError, parse_date
 import tools
 import wsgiref.util
-import threading
 import base64
 
 from bottle import BaseRequest, BaseResponse, LocalRequest
 
 class TestRequest(unittest.TestCase):
 
-    def test_app(self):
+    def test_app_property(self):
         e = {}
         r = BaseRequest(e)
         self.assertRaises(RuntimeError, lambda: r.app)
         e.update({'bottle.app': 5})
         self.assertEqual(r.app, 5)
 
+    def test_route_property(self):
+        e = {'bottle.route': 5}
+        r = BaseRequest(e)
+        self.assertEqual(r.route, 5)
+
+    def test_url_for_property(self):
+        e = {}
+        r = BaseRequest(e)
+        self.assertRaises(RuntimeError, lambda: r.url_args)
+        e.update({'route.url_args': {'a': 5}})
+        self.assertEqual(r.url_args, {'a': 5})
+
     def test_path(self):
         """ PATH_INFO normalization. """
         # Legal paths
@@ -268,6 +279,40 @@ class TestRequest(unittest.TestCase):
         self.assertEqual(42, len(request.body.readline()))
         self.assertEqual(42, len(request.body.readline(1024)))
 
+    def _test_chunked(self, body, expect):
+        e = {}
+        wsgiref.util.setup_testing_defaults(e)
+        e['wsgi.input'].write(tob(body))
+        e['wsgi.input'].seek(0)
+        e['HTTP_TRANSFER_ENCODING'] = 'chunked'
+        if isinstance(expect, str):
+            self.assertEquals(tob(expect), BaseRequest(e).body.read())
+        else:
+            self.assertRaises(expect, lambda: BaseRequest(e).body)
+
+    def test_chunked(self):
+        self._test_chunked('1\r\nx\r\nff\r\n' + 'y'*255 + '\r\n0\r\n',
+                           'x' + 'y'*255)
+        self._test_chunked('8\r\nxxxxxxxx\r\n0\r\n','xxxxxxxx')
+        self._test_chunked('0\r\n', '')
+
+    def test_chunked_meta_fields(self):
+        self._test_chunked('8 ; foo\r\nxxxxxxxx\r\n0\r\n','xxxxxxxx')
+        self._test_chunked('8;foo\r\nxxxxxxxx\r\n0\r\n','xxxxxxxx')
+        self._test_chunked('8;foo=bar\r\nxxxxxxxx\r\n0\r\n','xxxxxxxx')
+
+    def test_chunked_not_terminated(self):
+        self._test_chunked('1\r\nx\r\n', HTTPError)
+
+    def test_chunked_wrong_size(self):
+        self._test_chunked('2\r\nx\r\n', HTTPError)
+
+    def test_chunked_illegal_size(self):
+        self._test_chunked('x\r\nx\r\n', HTTPError)
+
+    def test_chunked_not_chunked_at_all(self):
+        self._test_chunked('abcdef', HTTPError)
+
     def test_multipart(self):
         """ Environ: POST (multipart files and multible values per key) """
         fields = [('field1','value1'), ('field2','value2'), ('field2','value3')]
@@ -328,7 +373,7 @@ class TestRequest(unittest.TestCase):
         e['wsgi.input'].write(tob(json_dumps(test)))
         e['wsgi.input'].seek(0)
         e['CONTENT_LENGTH'] = str(len(json_dumps(test)))
-        self.assertEqual(BaseRequest(e).json, None)
+        self.assertRaises(HTTPError, lambda: BaseRequest(e).json)
 
     def test_json_valid(self):
         """ Environ: Request.json property. """
@@ -340,15 +385,6 @@ class TestRequest(unittest.TestCase):
         e['CONTENT_LENGTH'] = str(len(json_dumps(test)))
         self.assertEqual(BaseRequest(e).json, test)
 
-    def test_json_forged_header_issue616(self):
-        test = dict(a=5, b='test', c=[1,2,3])
-        e = {'CONTENT_TYPE': 'text/plain;application/json'}
-        wsgiref.util.setup_testing_defaults(e)
-        e['wsgi.input'].write(tob(json_dumps(test)))
-        e['wsgi.input'].seek(0)
-        e['CONTENT_LENGTH'] = str(len(json_dumps(test)))
-        self.assertEqual(BaseRequest(e).json, None)
-
     def test_isajax(self):
         e = {}
         wsgiref.util.setup_testing_defaults(e)
@@ -390,27 +426,6 @@ class TestRequest(unittest.TestCase):
         del r.environ['HTTP_X_FORWARDED_FOR']
         self.assertEqual(r.remote_addr, ips[1])
 
-    def test_maxparam(self):
-        ips = ['1.2.3.4', '2.3.4.5', '3.4.5.6']
-        e = {}
-        wsgiref.util.setup_testing_defaults(e)
-        e['wsgi.input'].write(tob('a=a&b=b&c=c'))
-        e['wsgi.input'].seek(0)
-        e['CONTENT_LENGTH'] = '11'
-        e['REQUEST_METHOD'] = "POST"
-        e['HTTP_COOKIE'] = 'a=1,b=1,c=1;d=1'
-        e['QUERY_STRING'] = 'a&b&c&d'
-        old_value = BaseRequest.MAX_PARAMS
-        r = BaseRequest(e)
-        try:
-            BaseRequest.MAX_PARAMS = 2
-            self.assertEqual(len(list(r.query.allitems())), 2)
-            self.assertEqual(len(list(r.cookies.allitems())), 2)
-            self.assertEqual(len(list(r.forms.allitems())), 2)
-            self.assertEqual(len(list(r.params.allitems())), 4)
-        finally:
-            BaseRequest.MAX_PARAMS = old_value
-
     def test_user_defined_attributes(self):
         for cls in (BaseRequest, LocalRequest):
             r = cls()
@@ -427,6 +442,40 @@ class TestRequest(unittest.TestCase):
 
 class TestResponse(unittest.TestCase):
 
+    def test_constructor_body(self):
+        self.assertEqual('',
+            BaseResponse('').body)
+
+        self.assertEqual('YAY',
+            BaseResponse('YAY').body)
+
+    def test_constructor_status(self):
+        self.assertEqual(200,
+            BaseResponse('YAY', 200).status_code)
+
+        self.assertEqual('200 OK',
+            BaseResponse('YAY', 200).status_line)
+
+        self.assertEqual('200 YAY',
+            BaseResponse('YAY', '200 YAY').status_line)
+
+        self.assertEqual('200 YAY',
+            BaseResponse('YAY', '200 YAY').status_line)
+
+    def test_constructor_headerlist(self):
+        from functools import partial
+        make_res = partial(BaseResponse, '', 200)
+
+        self.assertTrue('yay',
+            make_res([('x-test','yay')])['x-test'])
+
+    def test_constructor_headerlist(self):
+        from functools import partial
+        make_res = partial(BaseResponse, '', 200)
+
+        self.assertTrue('yay', make_res(x_test='yay')['x-test'])
+
+
     def test_set_status(self):
         rs = BaseResponse()
 
@@ -589,7 +638,20 @@ class TestResponse(unittest.TestCase):
         response['x-test'] = None
         self.assertEqual('None', response['x-test'])
 
-
+    def test_expires_header(self):
+        import datetime
+        response = BaseResponse()
+        now = datetime.datetime.now()
+        response.expires = now
+        
+        def seconds(a, b):
+            td = max(a,b) - min(a,b)
+            return td.days*360*24 + td.seconds
+        
+        self.assertEqual(0, seconds(response.expires, now))
+        now2 = datetime.datetime.utcfromtimestamp(
+            parse_date(response.headers['Expires']))
+        self.assertEqual(0, seconds(now, now2))
 
 class TestRedirect(unittest.TestCase):
 
@@ -601,6 +663,7 @@ class TestRedirect(unittest.TestCase):
                 del args[key]
         env.update(args)
         request.bind(env)
+        bottle.response.bind()
         try:
             bottle.redirect(target, **(query or {}))
         except bottle.HTTPResponse:
@@ -687,6 +750,17 @@ class TestRedirect(unittest.TestCase):
                             'http://example.com/a%20a/b%20b/te st.html',
                             HTTP_HOST='example.com', SCRIPT_NAME='/a a/', PATH_INFO='/b b/')
 
+    def test_redirect_preserve_cookies(self):
+        env = {'SERVER_PROTOCOL':'HTTP/1.1'}
+        request.bind(env)
+        bottle.response.bind()
+        try:
+            bottle.response.set_cookie('xxx', 'yyy')
+            bottle.redirect('...')
+        except bottle.HTTPResponse:
+            h = [v for (k, v) in _e().headerlist if k == 'Set-Cookie']
+            self.assertEqual(h, ['xxx=yyy'])
+
 class TestWSGIHeaderDict(unittest.TestCase):
     def setUp(self):
         self.env = {}
